"""
Client for led service. Utilizes ClientRpcChannel to send/receive and
serialize/deserialize messages
"""
from rpc_generated_protobufs import led_pb2 as led_pb
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.services.led.led_pb_enums import LEDPins
from edgepi_rpc_client.util.helpers import get_server_response

# pylint: disable=no-member
class ClientLEDService():
    """Client methods for LED service"""
    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = led_pb.LEDService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    def turn_led_on(self, led_pin: LEDPins):
        """turn_on led method for sdk led module"""
        request = led_pb.LEDPin(
            led_pin = led_pin.value
        )
        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.turn_led_on(self.rpc_controller,request)

        response = get_server_response(rpc_response, led_pb.SuccessMsg)

        return response.content

    def turn_led_off(self, led_pin: LEDPins):
        """turn_off led method for sdk led module"""
        request = led_pb.LEDPin(
            led_pin = led_pin.value
        )
        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.turn_led_off(self.rpc_controller,request)

        response = get_server_response(rpc_response, led_pb.SuccessMsg)

        return response.content

    def toggle_led(self, led_pin: LEDPins):
        """toggle_led method for sdk led module"""
        request = led_pb.LEDPin(
            led_pin = led_pin.value
        )
        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.toggle_led(self.rpc_controller,request)

        response = get_server_response(rpc_response, led_pb.SuccessMsg)

        return response.content



"""Client enums to protobuf adc enums"""
from enum import Enum
from rpc_generated_protobufs import adc_pb2 as adc_pb

# pylint: disable=no-member
class AnalogIn(Enum):
    """AnalogIn Enum"""
    AIN1 = adc_pb.AnalogIn.AIN1
    AIN2 = adc_pb.AnalogIn.AIN2
    AIN3 = adc_pb.AnalogIn.AIN3
    AIN4 = adc_pb.AnalogIn.AIN4
    AIN5 = adc_pb.AnalogIn.AIN5
    AIN6 = adc_pb.AnalogIn.AIN6
    AIN7 = adc_pb.AnalogIn.AIN7
    AIN8 = adc_pb.AnalogIn.AIN8
    AINCOM = adc_pb.AnalogIn.AINCOM
    FLOAT = adc_pb.AnalogIn.FLOAT

class ADC1DataRate(Enum):
    """ADC1DataRate Enum"""
    SPS_2P5 = adc_pb.ADC1DataRate.SPS_2P5
    SPS_5 = adc_pb.ADC1DataRate.SPS_5
    SPS_10_ = adc_pb.ADC1DataRate.SPS_10_
    SPS_16P6 = adc_pb.ADC1DataRate.SPS_16P6
    SPS_20 = adc_pb.ADC1DataRate.SPS_20
    SPS_50 = adc_pb.ADC1DataRate.SPS_50
    SPS_60 = adc_pb.ADC1DataRate.SPS_60
    SPS_100_ = adc_pb.ADC1DataRate.SPS_100_
    SPS_400_ = adc_pb.ADC1DataRate.SPS_400_
    SPS_1200 = adc_pb.ADC1DataRate.SPS_1200
    SPS_2400 = adc_pb.ADC1DataRate.SPS_2400
    SPS_4800 = adc_pb.ADC1DataRate.SPS_4800
    SPS_7200 = adc_pb.ADC1DataRate.SPS_7200
    SPS_14400 = adc_pb.ADC1DataRate.SPS_14400
    SPS_19200 = adc_pb.ADC1DataRate.SPS_19200
    SPS_38400 = adc_pb.ADC1DataRate.SPS_38400

class ADC2DataRate(Enum):
    """ADC2DataRate Enum"""
    SPS_10 = adc_pb.ADC2DataRate.SPS_10
    SPS_100 = adc_pb.ADC2DataRate.SPS_100
    SPS_400 = adc_pb.ADC2DataRate.SPS_400
    SPS_800 = adc_pb.ADC2DataRate.SPS_800

class FilterMode(Enum):
    """FilterMode Enum"""
    SINC1 = adc_pb.FilterMode.SINC1
    SINC2 = adc_pb.FilterMode.SINC2
    SINC3 = adc_pb.FilterMode.SINC3
    SINC4 = adc_pb.FilterMode.SINC4
    FIR = adc_pb.FilterMode.FIR

class ConvMode(Enum):
    """ConvMode Enum"""
    CONTINUOUS = adc_pb.ConvMode.CONTINUOUS
    PULSE = adc_pb.ConvMode.PULSE

class ADCNum(Enum):
    """ADCNum Enum"""
    ADC_1 = adc_pb.ADCNum.ADC_1
    ADC_2 = adc_pb.ADCNum.ADC_2

class DiffMode(Enum):
    """DiffMode Enum"""
    DIFF_1 = adc_pb.DiffMode.DIFF_1
    DIFF_2 = adc_pb.DiffMode.DIFF_2
    DIFF_3 = adc_pb.DiffMode.DIFF_3
    DIFF_4 = adc_pb.DiffMode.DIFF_4
    DIFF_OFF = adc_pb.DiffMode.DIFF_OFF
    


"""LEDService integration tests"""
import pytest
from edgepi_rpc_client.services.led.client_led_service import ClientLEDService
from edgepi_rpc_client.services.led.led_pb_enums import LEDPins

@pytest.fixture(name='led_service')
def fixture_test_led_service():
    """Inits new led serviceclient for testing"""
    return ClientLEDService('tcp://localhost:5555')

@pytest.mark.parametrize(
    'led_pin',
    [
        (LEDPins.LED1),
        (LEDPins.LED2),
        (LEDPins.LED3),
        (LEDPins.LED4),
        (LEDPins.LED5),
        (LEDPins.LED6),
        (LEDPins.LED7),
        (LEDPins.LED8)
    ]
)
def test_toggle_led(led_service, led_pin):
    """Test for toggle_led"""
    response = led_service.toggle_led(led_pin)

    assert response == f'Successfully toggled {led_pin} to the opposite state.'

@pytest.mark.parametrize(
    'led_pin',
    [
        (LEDPins.LED1),
        (LEDPins.LED2),
        (LEDPins.LED3),
        (LEDPins.LED4),
        (LEDPins.LED5),
        (LEDPins.LED6),
        (LEDPins.LED7),
        (LEDPins.LED8)
    ]
)
def test_turn_led_on(led_service, led_pin):
    """Test for turn_led_on"""
    response = led_service.turn_led_on(led_pin)

    assert response == f'Successfully turned on {led_pin}.'

@pytest.mark.parametrize(
    'led_pin',
    [
        (LEDPins.LED1),
        (LEDPins.LED2),
        (LEDPins.LED3),
        (LEDPins.LED4),
        (LEDPins.LED5),
        (LEDPins.LED6),
        (LEDPins.LED7),
        (LEDPins.LED8)
    ]
)
def test_turn_led_off(led_service, led_pin):
    """Test for turn_led_off"""
    response = led_service.turn_led_off(led_pin)

    assert response == f'Successfully turned off {led_pin}.'



"""
General helper functions. If more helper functions are introducted,
helpers should be split up into different files by their category.
"""
from enum import Enum
import logging
from edgepi_rpc_client.error.rpc_failure import RpcFailure

_logger = logging.getLogger(__name__)

def create_config_arg_msg(arg_msg, arg_name, arg_value):
    """Creates a config protobuf message with the config argument name and value"""
    # Do not access enum value if it's the argument is not an enum
    setattr(
        arg_msg, arg_name, arg_value if not isinstance(arg_value,Enum) else \
        arg_value.value
    )
    return arg_msg

def create_config_request_from_args(config_msg, arg_msg, config_args_dict):
    """Creates a config message request with a given config argument dictionary"""
    # Create the set_config request message (Message with repeated arguments)
    # Append each config argument message to the config message
    for arg_name,arg_value in config_args_dict.items():
        request_argument = create_config_arg_msg(arg_msg,arg_name,arg_value)
        _logger.debug("Config argument message: %s", request_argument)
        config_msg.conf_arg.append(request_argument)

    return config_msg

def filter_arg_values(dictionary, filter_key, filter_value):
    """
    Gets a dictionary of arguments and filters unwanted keys by key or value. Returns a new
    dictionary
    """
    filtered_args_list = {
        key:value for (key,value) in dictionary.items()
                        if key != filter_key and value != filter_value
    }
    return filtered_args_list


def get_server_response(rpc_response, server_response_class):
    """Gets a response from the RPC Server"""
    # Check for error message
    if(rpc_response.HasField('error_msg') and rpc_response.HasField('error_code')):
        raise RpcFailure(f'{rpc_response.error_code}: {rpc_response.error_msg}')

    # Get server response from rpc message and deserialize
    server_response_data = rpc_response.response_proto
    server_response = server_response_class.FromString(server_response_data)

    return server_response



"""Client enums to protobuf PWM enums"""
from enum import Enum
from rpc_generated_protobufs import pwm_pb2 as pwm_pb

# pylint: disable=no-member
class PWMPins(Enum):
    """PWMPins Enum"""
    PWM1 = pwm_pb.PWMPins.PWM1
    PWM2 = pwm_pb.PWMPins.PWM2

class Polarity(Enum):
    """Polarity Enum"""
    NORMAL = pwm_pb.Polarity.NORMAL
    INVERSED = pwm_pb.Polarity.INVERSED



"""Client Enum to Protobuf Enum mapping"""
from enum import Enum, unique
from rpc_generated_protobufs import din_pb2 as din_pb

# pylint: disable=no-member
@unique
class DinPins(Enum):
    """DinPins Enum"""
    DIN1 = din_pb.DinPins.DIN1
    DIN2 = din_pb.DinPins.DIN2
    DIN3 = din_pb.DinPins.DIN3
    DIN4 = din_pb.DinPins.DIN4
    DIN5 = din_pb.DinPins.DIN5
    DIN6 = din_pb.DinPins.DIN6
    DIN7 = din_pb.DinPins.DIN7
    DIN8 = din_pb.DinPins.DIN8



"""
Relay service integration tests
"""
import pytest
from edgepi_rpc_client.services.relay.client_relay_service import ClientRelayService

@pytest.fixture(name='relay_service')
def fixture_test_relay_service():
    """Inits new relay service client for testing"""
    return ClientRelayService('tcp://localhost:5555')

def test_open_relay(relay_service):
    """Test client call of open_relay and compare with client call of get_state_relay"""
    # Open relay
    response = relay_service.open_relay()

    state = relay_service.get_state_relay()

    assert not state
    assert response == "Successfully opened relay"

def test_close_relay(relay_service):
    """Test client call of close_relay and compare with client call of get_state_relay"""
    # Open relay
    response = relay_service.close_relay()

    state = relay_service.get_state_relay()

    assert state
    assert response == "Successfully closed relay"



"""
Client for Din service. Utilizes ClientRpcChannel to send/receive and
serialize/deserialize messages
"""
from rpc_generated_protobufs import din_pb2 as din_pb
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.services.din.din_pb_enums import DinPins
from edgepi_rpc_client.util.helpers import get_server_response

# pylint: disable=no-member
# pylint: disable=too-few-public-methods
class ClientDinService():
    """Client methods for Din Service"""
    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = din_pb.DinService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    def digital_input_state(self, din_pin: DinPins):
        """digital_input_state method for sdk din module"""
        request = din_pb.DinPin(
            din_pin = din_pin.value
        )
        # Call SDK method through rpc channel
        rpc_response = self.service_stub.digital_input_state(self.rpc_controller, request)

        response = get_server_response(rpc_response, din_pb.State)

        return response.state_bool



'''Client RPC Channel implementation'''

import logging
from google.protobuf import service
import zmq
from rpc_generated_protobufs import rpc_pb2 as rpc_pb

_logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.DEBUG
)

# pylint: disable=too-few-public-methods
class ClientRpcChannel(service.RpcChannel):
    """
    Client class that utilizes google's abstract RpcChannel interfce. 
    - The RpcChannel represents a comunication line to a service and can 
    be used to call that service's methods. 
    - This class handles connecting to the RPC server, as well as serializing/deserializing
    messages to and from the server when service methods are called from the client side.
    """

    def __init__(self, transport):
        self.transport = transport
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect(self.transport)

    # pylint: disable=no-member
    def _get_rpc_request(self, method, client_request):
        # Wrap client's request msg in an RPC message
        rpc_request = rpc_pb.RpcRequest()
        # Set rpc request to client's request as byte string
        rpc_request.request_proto = client_request.SerializeToString()
        # Set rpc service and method name
        rpc_request.service_name = method.containing_service.name
        rpc_request.method_name = method.name
        _logger.debug("RPC request object: {%s}", rpc_request)
        return rpc_request

    def _send_rpc_request(self, rpc_request):
        try:
            # Serialize RPC protocol message
            # Send over socket
            rpc_request_data = rpc_request.SerializeToString()
            self.socket.send(rpc_request_data)
        # pylint: disable=broad-except, broad-exception-caught
        except Exception as exc:
            _logger.exception("Error serialzing/sending RPC request to server: %s", exc)

    #pylint: disable=no-member
    def _get_rpc_response(self):
        # Receive rpc response data from socket
        try:
            rpc_response_data = self.socket.recv()
            # Return the rpc response message
            rpc_response = rpc_pb.RpcResponse()
            rpc_response = rpc_response.FromString(rpc_response_data)
            _logger.debug("RPC response object: {%s}", rpc_response)
            return rpc_response
        # pylint: disable=broad-exception-caught
        except Exception as exc:
            _logger.exception("Error recieving/deserializing RPC response: {%s}", exc)
        return None

    # pylint: disable=too-many-arguments, unused-argument
    # TODO: Determine whether unused arguments 'response_class' and 'done' are needed
    def CallMethod(self, method_descriptor, _rpc_controller, request, response_class, done):
        _logger.debug('%s', request)
        # Create rpc request
        rpc_request = self._get_rpc_request(method_descriptor,request)
        # Send rpc service request over socket
        self._send_rpc_request(rpc_request)
        # Get the rpc response from socket
        rpc_response = self._get_rpc_response()
        # Return rpc response message
        return rpc_response



"""
Client for Dac service. Utilizes ClientRpcChannel to send/receive and
serialize/deserialize messages
"""
from rpc_generated_protobufs import dac_pb2 as dac_pb
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.services.dac.dac_pb_enums import DACChannel
from edgepi_rpc_client.util.helpers import get_server_response

# pylint: disable=no-member
class ClientDacService():
    """Client methods for Dac Service"""
    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = dac_pb.DacService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    def set_dac_gain(self, set_gain: bool, auto_code_change: bool = False):
        """set_dac_gain method for sdk dac module"""
        request = dac_pb.Gain(
            set_gain=set_gain,
            auto_code_change=auto_code_change
        )

        # Call SDK method through rpc channel
        rpc_response = self.service_stub.set_dac_gain(self.rpc_controller, request)

        response = get_server_response(rpc_response, dac_pb.GainState)

        return response.gain_state

    def write_voltage(self, analog_out: DACChannel, voltage: float):
        """write_voltage method for sdk dac module"""
        request = dac_pb.WriteVoltage(
            dac_channel=analog_out.value,
            voltage=voltage
        )

        rpc_response = self.service_stub.write_voltage(self.rpc_controller, request)

        response = get_server_response(rpc_response, dac_pb.SuccessMsg)

        return response.content

    def get_state(self, analog_out: DACChannel = None, code: bool = None,
                        voltage: bool = None, gain: bool = None):
        """get_state method for sdk dac module"""
        request = dac_pb.GetState(
            dac_channel=analog_out.value,
            code=code,
            voltage=voltage,
            gain=gain
        )

        rpc_response = self.service_stub.get_state(self.rpc_controller, request)

        response = get_server_response(rpc_response, dac_pb.State)

        return response.code_val, response.voltage_val, response.gain_state

    def reset(self):
        """reset method for sdk dac module"""
        request = dac_pb.EmptyMsg()

        rpc_response = self.service_stub.reset(self.rpc_controller, request)

        response = get_server_response(rpc_response, dac_pb.SuccessMsg)

        return response.content



"""Client enums to protobuf led enums"""
# pylint: disable=no-member
from enum import Enum, unique
from rpc_generated_protobufs import dout_pb2 as dout_pb

@unique
class DoutPins(Enum):
    """DoutPins Enum"""
    DOUT1 = dout_pb.DoutPins.DOUT1
    DOUT2 = dout_pb.DoutPins.DOUT2
    DOUT3 = dout_pb.DoutPins.DOUT3
    DOUT4 = dout_pb.DoutPins.DOUT4
    DOUT5 = dout_pb.DoutPins.DOUT5
    DOUT6 = dout_pb.DoutPins.DOUT6
    DOUT7 = dout_pb.DoutPins.DOUT7
    DOUT8 = dout_pb.DoutPins.DOUT8

@unique
class DoutTriState(Enum):
    """DoutPins Enum"""
    HI_Z = dout_pb.DoutTriState.HI_Z
    HIGH = dout_pb.DoutTriState.HIGH
    LOW = dout_pb.DoutTriState.LOW



"""
DoutService integration tests. TODO: 
use get_state rpc service method once implement to compare expected resulst instead of edgepi sdk
"""
import pytest
from edgepi.digital_output.edgepi_digital_output import EdgePiDigitalOutput
from edgepi.digital_output.digital_output_constants import (
    DoutPins as sdkDoutPins, DoutTriState as sdkDoutTriState)
from edgepi_rpc_client.services.dout.client_dout_service import ClientDoutService
from edgepi_rpc_client.services.dout.dout_pb_enums import DoutPins,DoutTriState

@pytest.fixture(name='dout_service')
def fixture_test_dout_service():
    """Inits new led serviceclient for testing"""
    return ClientDoutService('tcp://localhost:5555')

@pytest.mark.parametrize("pin, sdk_pin", [
    (DoutPins.DOUT1, sdkDoutPins.DOUT1),
    (DoutPins.DOUT2, sdkDoutPins.DOUT2),
    (DoutPins.DOUT3, sdkDoutPins.DOUT3),
    (DoutPins.DOUT4, sdkDoutPins.DOUT4),
    (DoutPins.DOUT5, sdkDoutPins.DOUT5),
    (DoutPins.DOUT6, sdkDoutPins.DOUT6),
    (DoutPins.DOUT7, sdkDoutPins.DOUT7),
    (DoutPins.DOUT8, sdkDoutPins.DOUT8),
])
def test_output_high(dout_service,pin, sdk_pin):
    """Test client call for set_dout_state and compares results using sdk method and enum"""
    edgepi_dout = EdgePiDigitalOutput()
    client_dout = dout_service
    response =client_dout.set_dout_state(pin, DoutTriState.HIGH)
    dout_state = edgepi_dout.get_state(sdk_pin)
    assert dout_state == sdkDoutTriState.HIGH
    assert response == f'Successfully set {sdk_pin} to DoutTriState.HIGH.'

@pytest.mark.parametrize("pin, sdk_pin", [
    (DoutPins.DOUT1, sdkDoutPins.DOUT1),
    (DoutPins.DOUT2, sdkDoutPins.DOUT2),
    (DoutPins.DOUT3, sdkDoutPins.DOUT3),
    (DoutPins.DOUT4, sdkDoutPins.DOUT4),
    (DoutPins.DOUT5, sdkDoutPins.DOUT5),
    (DoutPins.DOUT6, sdkDoutPins.DOUT6),
    (DoutPins.DOUT7, sdkDoutPins.DOUT7),
    (DoutPins.DOUT8, sdkDoutPins.DOUT8),
])
def test_output_z(dout_service,pin, sdk_pin):
    """Test client call for set_dout_state and compares results using sdk method and enum"""
    edgepi_dout = EdgePiDigitalOutput()
    client_dout = dout_service
    response = client_dout.set_dout_state(pin, DoutTriState.HI_Z)
    dout_state = edgepi_dout.get_state(sdk_pin)
    assert dout_state == sdkDoutTriState.HI_Z
    assert response == f'Successfully set {sdk_pin} to DoutTriState.HI_Z.'

@pytest.mark.parametrize("pin, sdk_pin", [
    (DoutPins.DOUT1, sdkDoutPins.DOUT1),
    (DoutPins.DOUT2, sdkDoutPins.DOUT2),
    (DoutPins.DOUT3, sdkDoutPins.DOUT3),
    (DoutPins.DOUT4, sdkDoutPins.DOUT4),
    (DoutPins.DOUT5, sdkDoutPins.DOUT5),
    (DoutPins.DOUT6, sdkDoutPins.DOUT6),
    (DoutPins.DOUT7, sdkDoutPins.DOUT7),
    (DoutPins.DOUT8, sdkDoutPins.DOUT8),
])
def test_output_low(dout_service,pin, sdk_pin):
    """Test client call for set_dout_state and compares results using sdk method and enum"""
    edgepi_dout = EdgePiDigitalOutput()
    client_dout = dout_service
    response = client_dout.set_dout_state(pin, DoutTriState.LOW)
    dout_state = edgepi_dout.get_state(sdk_pin)
    assert dout_state == sdkDoutTriState.LOW
    assert response == f'Successfully set {sdk_pin} to DoutTriState.LOW.'



"""Client enums to protobuf led enums"""
# pylint: disable=no-member
from enum import Enum, unique
from rpc_generated_protobufs import led_pb2 as led_pb

@unique
class LEDPins(Enum):
    """LEDPins Enum"""
    LED1 = led_pb.LEDPins.LED1
    LED2 = led_pb.LEDPins.LED2
    LED3 = led_pb.LEDPins.LED3
    LED4 = led_pb.LEDPins.LED4
    LED5 = led_pb.LEDPins.LED5
    LED6 = led_pb.LEDPins.LED6
    LED7 = led_pb.LEDPins.LED7
    LED8 = led_pb.LEDPins.LED8



"""DinService integration tests"""
import pytest
from edgepi.digital_input.edgepi_digital_input import EdgePiDigitalInput
from edgepi.digital_input.digital_input_constants import DinPins as sdkDinPins
from edgepi_rpc_client.services.din.client_din_service import ClientDinService
from edgepi_rpc_client.services.din.din_pb_enums import DinPins

@pytest.fixture(name='din_service')
def fixture_test_din_service():
    """Inits new din service client for testing"""
    return ClientDinService('tcp://localhost:5555')

@pytest.mark.parametrize("pin, sdk_pin", [
    (DinPins.DIN1, sdkDinPins.DIN1),
    (DinPins.DIN2, sdkDinPins.DIN2),
    (DinPins.DIN3, sdkDinPins.DIN3),
    (DinPins.DIN4, sdkDinPins.DIN4),
    (DinPins.DIN5, sdkDinPins.DIN5),
    (DinPins.DIN6, sdkDinPins.DIN6),
    (DinPins.DIN7, sdkDinPins.DIN7),
    (DinPins.DIN8, sdkDinPins.DIN8),
])
def test_digital_input_state(din_service, pin, sdk_pin):
    """Test client call for set_dout_state and compares results using sdk method and enum"""
    edgepi_din = EdgePiDigitalInput()
    client_din = din_service
    response =client_din.digital_input_state(pin)
    din_state = edgepi_din.digital_input_state(sdk_pin)
    assert response == din_state



"""
Client for Dout service. Utilizes ClientRpcChannel to send/receive and
serialize/deserialize messages
"""
from rpc_generated_protobufs import dout_pb2 as dout_pb
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.services.dout.dout_pb_enums import DoutPins, DoutTriState
from edgepi_rpc_client.util.helpers import get_server_response

# pylint: disable=no-member, too-few-public-methods
class ClientDoutService():
    """Client methods for Dout service"""
    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = dout_pb.DoutService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    def set_dout_state(self, dout_pin: DoutPins, state:DoutTriState):
        """set_dout_state method for sdk dout module"""
        request = dout_pb.PinAndState(
            dout_pin = dout_pin.value,
            state = state.value
        )
        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.set_dout_state(self.rpc_controller,request)

        response = get_server_response(rpc_response, dout_pb.SuccessMsg)

        return response.content



"""Client for PWM service"""
import logging
from rpc_generated_protobufs import pwm_pb2 as pwm_pb
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.util.helpers import (
    filter_arg_values,
    create_config_request_from_args,
    get_server_response,
)
from edgepi_rpc_client.services.pwm.pwm_pb_enums import PWMPins, Polarity

_logger = logging.getLogger(__name__)


# pylint: disable=no-member
class ClientPWMService:
    """Client Methods for PWM Service"""

    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = pwm_pb.PWMService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    # pylint: disable=unused-argument
    def set_config(
        self,
        pwm_num: PWMPins,
        frequency: float = None,
        duty_cycle: float = None,
        polarity: Polarity = None,
    ):
        """set_config method for SDK PWM module"""
        config_args_dict = filter_arg_values(locals(), "self", None)
        config_msg = pwm_pb.Config()
        arg_msg = pwm_pb.Config().ConfArg()
        request = create_config_request_from_args(config_msg, arg_msg, config_args_dict)
        return self.perform_call_from_request(
            request, self.service_stub.set_config, pwm_pb.SuccessMsg
        ).content

    def enable(self, pwm_num: PWMPins):
        """enable method for SDK PWM module"""
        return self.perform_rpc_call(
            pwm_num, self.service_stub.enable, pwm_pb.SuccessMsg
        ).content

    def disable(self, pwm_num: PWMPins):
        """disable method for SDK PWM module"""
        return self.perform_rpc_call(
            pwm_num, self.service_stub.disable, pwm_pb.SuccessMsg
        ).content

    def close(self, pwm_num: PWMPins):
        """close method for SDK PWM module"""
        return self.perform_rpc_call(
            pwm_num, self.service_stub.close, pwm_pb.SuccessMsg
        ).content

    def init_pwm(self, pwm_num: PWMPins):
        """init_pwm method for SDK PWM module"""
        return self.perform_rpc_call(
            pwm_num, self.service_stub.init_pwm, pwm_pb.SuccessMsg
        ).content

    def get_frequency(self, pwm_num: PWMPins):
        """get_frequency method for SDK PWM module"""
        return self.perform_rpc_call(
            pwm_num, self.service_stub.get_frequency, pwm_pb.GetFrequency
        ).frequency

    def get_duty_cycle(self, pwm_num: PWMPins):
        """get_duty_cycle method for SDK PWM module"""
        return self.perform_rpc_call(
            pwm_num, self.service_stub.get_duty_cycle, pwm_pb.GetDutyCycle
        ).duty_cycle

    def get_polarity(self, pwm_num: PWMPins):
        """get_polarity method for SDK PWM module"""
        polarity = self.perform_rpc_call(
            pwm_num, self.service_stub.get_polarity, pwm_pb.GetPolarity
        ).polarity
        return Polarity(polarity)

    def get_enabled(self, pwm_num: PWMPins):
        """get_enabled method for SDK PWM module"""
        return self.perform_rpc_call(
            pwm_num, self.service_stub.get_enabled, pwm_pb.GetEnabled
        ).enabled

    # TODO: Potentially these helpers in other services and put them in a separate file
    def perform_rpc_call(self, pwm_num, method, response_type):
        """Performs RPC call with PWM number and specified method"""
        request = pwm_pb.PWM(pwm_num=pwm_num.value)
        return self.perform_call_from_request(request, method, response_type)

    def perform_call_from_request(self, request, method, response_type):
        """Executes RPC call using provided request and method"""
        rpc_response = method(self.rpc_controller, request)
        return get_server_response(rpc_response, response_type)



"""
Client for Relay Service
"""
from rpc_generated_protobufs import relay_pb2 as relay_pb
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.util.helpers import get_server_response


# pylint: disable=no-member
class ClientRelayService():
    """Client methods for relay service"""
    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = relay_pb.RelayService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    def open_relay(self):
        """open_relay method for sdk relay module"""
        request = relay_pb.EmptyMsg()

        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.open_relay(self.rpc_controller,request)

        response = get_server_response(rpc_response, relay_pb.SuccessMsg)

        return response.content

    def close_relay(self):
        """close_relay method for sdk relay module"""
        request = relay_pb.EmptyMsg()

        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.close_relay(self.rpc_controller,request)

        response = get_server_response(rpc_response, relay_pb.SuccessMsg)

        return response.content

    def get_state_relay(self):
        """get_state_relay method for sdk relay module"""
        request = relay_pb.EmptyMsg()

        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.get_state_relay(self.rpc_controller,request)

        response = get_server_response(rpc_response, relay_pb.State)

        return response.state_bool



"""Integration tests DAC service"""
import pytest
from edgepi.dac.edgepi_dac import EdgePiDAC
from edgepi_rpc_client.services.dac.client_dac_service import ClientDacService
from edgepi_rpc_client.services.dac.dac_pb_enums import DACChannel

@pytest.fixture(name="dac_service")
def fixture_test_dac_service():
    """Inits new dac service client for testing"""
    return ClientDacService('tcp://localhost:5555')


def test_reset(dac_service):
    """Tests client call for dac reset"""
    response = dac_service.reset()

    assert response == "Successfully reset DAC."

@pytest.mark.parametrize("channel, voltage,", [
    (DACChannel.AOUT1, 3.0),
    (DACChannel.AOUT2, 3.0),
    (DACChannel.AOUT3, 3.0),
    (DACChannel.AOUT4, 3.0),
    (DACChannel.AOUT5, 3.0),
    (DACChannel.AOUT6, 3.0),
    (DACChannel.AOUT7, 3.0),
    (DACChannel.AOUT8, 3.0)
])
def test_write_read_voltage(dac_service, channel, voltage):
    """Test client call for write voltage and compare with result of client call of get_state"""

    write_response = dac_service.write_voltage(channel,voltage)

    assert write_response == f"Successfully wrote {voltage}v to {channel}."

    _,read_voltage,_ = dac_service.get_state(channel, False, True, False)

    # assert read_voltage is voltage
    assert read_voltage == pytest.approx(voltage, abs=1e-3)

    # Reset
    dac_service.reset()

@pytest.mark.parametrize("gain", [
    (True),
    (False)
])
def test_set_get_gain(dac_service, gain):
    """Test client call for setting gain and then compare with client call of getting gain"""

    gain_state = dac_service.set_dac_gain(gain)

    _,_,read_gain = dac_service.get_state(DACChannel.AOUT1, False, False, True)

    assert gain_state == gain and gain == read_gain

    dac_service.reset()

def test_code(dac_service):
    """Tests that auto_code_change functionality works and that reading code works"""
    # Init dac (sdk)
    edgepi_dac = EdgePiDAC()

    # Enforce gain and auto code change to false
    dac_service.set_dac_gain(False,False)

    # write voltage
    dac_service.write_voltage(DACChannel.AOUT1, 3.3)

    # change gain with auto_code_change enabled
    dac_service.set_dac_gain(True,True)

    # read code and voltage
    read_code,read_voltage,_ = dac_service.get_state(DACChannel.AOUT1, True, True, True)

    # check voltage stayed the same and code is correct
    assert read_voltage == pytest.approx(3.3, abs=1e-3)
    assert read_code == edgepi_dac.dac_ops.voltage_to_code(DACChannel.AOUT1.value, read_voltage, 2)

    # Reset
    dac_service.reset()



"""
Client for adc service
"""
import logging
from rpc_generated_protobufs import adc_pb2 as adc_pb
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.util.helpers import (
    filter_arg_values, create_config_request_from_args, get_server_response
)
from edgepi_rpc_client.services.adc.adc_pb_enums import (
    AnalogIn, ConvMode, ADC1DataRate, ADC2DataRate, FilterMode, ADCNum, DiffMode
)

_logger = logging.getLogger(__name__)

# pylint: disable=no-member
class ClientAdcService():
    """Client Methods for Adc Service"""
    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = adc_pb.AdcService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    # pylint: disable=unused-argument, too-many-arguments
    def set_config(self,
            adc_1_analog_in: AnalogIn = None,
            adc_1_data_rate: ADC1DataRate = None,
            adc_2_analog_in: AnalogIn = None,
            adc_2_data_rate: ADC2DataRate = None,
            filter_mode: FilterMode = None,
            conversion_mode: ConvMode = None,
            override_updates_validation: bool = False
        ):
        """set config method for sdk adc module"""
        # Get a dictionary of arguments that are not None.
        config_args_dict= filter_arg_values(locals(), 'self', None)
        _logger.debug("Config argument dictionary: %s", config_args_dict)

        # Create config request
        config_msg = adc_pb.Config()
        arg_msg = adc_pb.Config().ConfArg()
        request = create_config_request_from_args(config_msg, arg_msg, config_args_dict)

        # Call the SDK method through the rpc channel client
        rpc_response = self.service_stub.set_config(self.rpc_controller,request)

        response = get_server_response(rpc_response, adc_pb.SuccessMsg)

        return response.content

    def single_sample(self):
        """single_sample method for sdk adc module"""
        request = adc_pb.EmptyMsg()
        # call the SDK method through rpc channel client
        rpc_response = self.service_stub.single_sample(self.rpc_controller,request)

        response = get_server_response(rpc_response, adc_pb.VoltageRead)

        voltage = response.voltage_read

        return voltage

    def select_differential(self, adc: ADCNum, diff_mode: DiffMode ):
        """select_differential method for sdk adc module"""
        request = adc_pb.DiffConfig(
            adc_num = adc.value,
            diff = diff_mode.value
        )

        # call the SDK method through rpc channel client
        rpc_response = self.service_stub.select_differential(self.rpc_controller, request)

        response = get_server_response(rpc_response, adc_pb.SuccessMsg)

        return response.content

    def set_rtd(self, set_rtd: bool, adc: ADCNum = ADCNum.ADC_2):
        """set_rtd method for sdk adc module"""
        request = adc_pb.RtdConfig(
            set_rtd = set_rtd,
            adc_num = adc.value
        )

        # call the SDK method through rpc channel client
        rpc_response = self.service_stub.set_rtd(self.rpc_controller, request)

        response = get_server_response(rpc_response, adc_pb.SuccessMsg)

        return response.content

    def single_sample_rtd(self):
        """single_sample_rtd method for sdk adc module"""
        request = adc_pb.EmptyMsg()

        # call the SDK method through rpc channel client
        rpc_response = self.service_stub.single_sample_rtd(self.rpc_controller, request)

        response = get_server_response(rpc_response, adc_pb.TempReading)

        temp = response.temp

        return temp

    def start_conversions(self, adc_num: ADCNum):
        """start_conversions method for sdk adc module"""
        # call start or stop converions
        request = adc_pb.ADC(
            adc_num = adc_num.value
        )

        # call sdk method through rpc channel client
        rpc_response = self.service_stub.start_conversions(self.rpc_controller, request)

        response = get_server_response(rpc_response, adc_pb.SuccessMsg)

        return response.content

    def stop_conversions(self, adc_num: ADCNum):
        """stop_conversions method for sdk adc module"""
        # call start or stop converions
        request = adc_pb.ADC(
            adc_num = adc_num.value
        )

        # call sdk method through rpc channel client
        rpc_response = self.service_stub.stop_conversions(self.rpc_controller, request)

        response = get_server_response(rpc_response, adc_pb.SuccessMsg)

        return response.content

    def read_voltage(self, adc_num: ADCNum):
        """read_voltage method for sdk adc module"""
        request = adc_pb.ADC(
            adc_num = adc_num.value
        )

        # call the SDK method through rpc channel client
        rpc_response = self.service_stub.read_voltage(self.rpc_controller, request)

        response = get_server_response(rpc_response, adc_pb.VoltageRead)

        return response.voltage_read

    def read_rtd_temperature(self):
        """read_rtd_temperature method for sdk adc module"""
        request = adc_pb.EmptyMsg()

        # call the SDK method through rpc channel client
        rpc_response = self.service_stub.read_rtd_temperature(self.rpc_controller, request)

        response = get_server_response(rpc_response, adc_pb.TempReading)

        temp = response.temp

        return temp



"""AdcService integration test"""
import logging
import pytest
from edgepi.adc.edgepi_adc import EdgePiADC
from edgepi_rpc_client.services.adc.client_adc_service import ClientAdcService
from edgepi_rpc_client.services.adc.adc_pb_enums import (
    AnalogIn, ConvMode, ADC1DataRate, FilterMode, ADC2DataRate, ADCNum, DiffMode)

_logger = logging.getLogger(__name__)

@pytest.fixture(name="adc_service")
def fixture_test_tc_service():
    """Inits new tc service client for testing"""
    return ClientAdcService('tcp://localhost:5555')



@pytest.mark.parametrize(
    "args",
    [
        (
            {
                "adc_1_analog_in": AnalogIn.AIN1
            }
        ),
        (
            {
                "adc_1_analog_in": AnalogIn.AIN2
            }
        ),
        (
            {
                "adc_1_analog_in": AnalogIn.AIN3
            }
        ),
        (
            {
                "adc_1_analog_in": AnalogIn.AIN4
            }
        ),
        (
            {
                "adc_1_analog_in": AnalogIn.AIN5
            }
        ),
        (
            {
                "adc_1_analog_in": AnalogIn.AIN6
            }
        ),
        (
            {
                "adc_1_analog_in": AnalogIn.AIN7
            }
        ),
        (
            {
                "adc_1_analog_in": AnalogIn.AIN8
            }
        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN1
            }
        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN2
            }
        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN3
            }
        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN4
            }
        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN5
            }
        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN6
            }

        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN7
            }
        ),
        (
            {
                "adc_2_analog_in": AnalogIn.AIN8
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_2P5
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_5
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_10_
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_16P6
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_20
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_50
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_60
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_100_
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_400_
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_1200
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_2400
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_4800
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_7200
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_14400
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_19200
            }
        ),
        (
            {
                "adc_1_data_rate": ADC1DataRate.SPS_38400
            }
        ),
        (
            {
                "adc_2_data_rate": ADC2DataRate.SPS_10
            }
        ),
        (
            {
                "adc_2_data_rate": ADC2DataRate.SPS_100
            }
        ),
        (
            {
                "adc_2_data_rate": ADC2DataRate.SPS_400
            }
        ),
        (
            {
                "adc_2_data_rate": ADC2DataRate.SPS_800
            }
        ),
        (
            {
                "filter_mode": FilterMode.SINC1
            }
        ),
        (
            {
                "filter_mode": FilterMode.SINC2
            }
        ),
        (
            {
                "filter_mode": FilterMode.SINC3
            }
        ),
        (
            {
                "filter_mode": FilterMode.SINC4
            }
        ),
        (
            {
                "filter_mode": FilterMode.FIR
            }
        ),
        (
            {
                "conversion_mode": ConvMode.PULSE
            }
        ),
        (
            {
                "conversion_mode": ConvMode.CONTINUOUS
            }
        )
    ]
)
def test_set_config(adc_service, args):
    """Test for set_config ADC"""
    adc = EdgePiADC()

    response = adc_service.set_config(**args)
    assert response == "Successfully applied adc configurations using set_config"
    adc.reset()

def test_pulse_conversion_mode(adc_service):
    """"test for pulse conversion mode for adc service method"""

    # set config
    response = adc_service.set_config(adc_1_analog_in=AnalogIn.AIN1,
         conversion_mode=ConvMode.PULSE, adc_1_data_rate=ADC1DataRate.SPS_38400)

    # single sample
    out = adc_service.single_sample()

    _logger.debug('%s', out)

    assert response == "Successfully applied adc configurations using set_config"
    assert out != 0

@pytest.mark.parametrize(
    "diff_args, diff_str",
    [
        (
            {
                "adc": ADCNum.ADC_1,
                "diff_mode": DiffMode.DIFF_1
            },
            "DiffMode.DIFF_1"
        ),
        (
            {
                "adc": ADCNum.ADC_1,
                "diff_mode": DiffMode.DIFF_2
            },
            "DiffMode.DIFF_2"
        ),
        (
            {
                "adc": ADCNum.ADC_1,
                "diff_mode": DiffMode.DIFF_3
            },
            "DiffMode.DIFF_3"
        ),
        (
            {
                "adc": ADCNum.ADC_1,
                "diff_mode": DiffMode.DIFF_4
            },
            "DiffMode.DIFF_4"
        ),
        (
            {
                "adc": ADCNum.ADC_1,
                "diff_mode": DiffMode.DIFF_OFF
            },
            "DiffMode.DIFF_OFF"
        ),
        (
            {
                "adc": ADCNum.ADC_2,
                "diff_mode": DiffMode.DIFF_1
            },
            "DiffMode.DIFF_1"
        ),
        (
            {
                "adc": ADCNum.ADC_2,
                "diff_mode": DiffMode.DIFF_2
            },
            "DiffMode.DIFF_2"
        ),
        (
            {
                "adc": ADCNum.ADC_2,
                "diff_mode": DiffMode.DIFF_3
            },
            "DiffMode.DIFF_3"
        ),
        (
            {
                "adc": ADCNum.ADC_2,
                "diff_mode": DiffMode.DIFF_4
            },
            "DiffMode.DIFF_4"
        ),
        (
            {
                "adc": ADCNum.ADC_2,
                "diff_mode": DiffMode.DIFF_OFF
            },
            "DiffMode.DIFF_OFF"
        )
    ]
)
def test_continuous_differential(adc_service, diff_args, diff_str):
    """test differential functionality"""
    # config to pulse mode
    adc_service.set_config(conversion_mode=ConvMode.CONTINUOUS)

    # select differential
    diff_response = adc_service.select_differential(**diff_args)
    assert diff_response == f"Successfully selected {diff_str}."

    # don't continue on diff off
    if diff_args["diff_mode"] == DiffMode.DIFF_OFF:
        return

    # start conversions
    start_conv_response = adc_service.start_conversions(diff_args["adc"])
    assert start_conv_response == "Successfully started conversions."

    # read voltage
    voltage = adc_service.read_voltage(diff_args["adc"])
    assert isinstance(voltage, float)

    # stop conversions
    stop_conv_response = adc_service.stop_conversions(diff_args["adc"])
    assert stop_conv_response == "Successfully stopped conversions."

@pytest.mark.parametrize(
    "set_rtd_args, set_rtd_str",
    [
        (
            {
                "set_rtd": True,
                "adc": ADCNum.ADC_1
            },
            "on"
        ),
        (
            {
                "set_rtd": False,
                "adc": ADCNum.ADC_1
            },
            "off"
        ),
        (
            {
                "set_rtd": True,
                "adc": ADCNum.ADC_2
            },
            "on"
        ),
        (
            {
                "set_rtd": False,
                "adc": ADCNum.ADC_2
            },
            "off"
        ),
    ]
)
def test_rtd(adc_service, set_rtd_args, set_rtd_str):
    """test rtd functionaltiy"""
    # config to pulse mode
    adc_service.set_config(conversion_mode=ConvMode.PULSE)

    # set rtd
    response = adc_service.set_rtd(**set_rtd_args)
    assert response == f'Successfully set RTD {set_rtd_str}.'

    if set_rtd_args["set_rtd"] and set_rtd_args["adc"] == ADCNum.ADC_1:
        temp = adc_service.single_sample_rtd()
        assert isinstance(temp, float)


@pytest.mark.parametrize(
    "adc",
    [
        (ADCNum.ADC_1),(ADCNum.ADC_2),
    ]
)
def test_read_rtd_temperature(adc_service, adc):
    """test read_rtd_temperature functionality"""
    # config to continuous mode
    adc_service.set_config(conversion_mode=ConvMode.CONTINUOUS)

    # start conversions
    adc_service.start_conversions(adc)

    # set rtd conversions
    adc_service.set_rtd(True, adc)

    temp = adc_service.read_rtd_temperature()
    assert isinstance(temp, float)

    # stop conversions
    adc_service.stop_conversions(adc)

    # set rtd conversions
    adc_service.set_rtd(False, adc)



"""Client enums to protobuf tc enums"""
#pylint: disable=no-member
from enum import Enum, unique
from rpc_generated_protobufs import tc_pb2 as tc_pb

@unique
class ConvMode(Enum):
    """ConvMode Enum"""
    SINGLE = tc_pb.ConvMode.SINGLE
    AUTO = tc_pb.ConvMode.AUTO

@unique
class OpenCircuitMode(Enum):
    """OpenCircuitMode Enum"""
    DISABLED = tc_pb.OpenCircuitMode.DISABLED
    LOW_INPUT_IMPEDANCE = tc_pb.OpenCircuitMode.LOW_INPUT_IMPEDANCE
    MED_INPUT_IMPEDANCE = tc_pb.OpenCircuitMode.MED_INPUT_IMPEDANCE
    HIGH_INPUT_IMPEDANCE = tc_pb.OpenCircuitMode.HIGH_INPUT_IMPEDANCE

@unique
class CJMode(Enum):
    """CJMode Eum"""
    ENABLE = tc_pb.CJMode.ENABLE
    DISABLE = tc_pb.CJMode.DISABLE

@unique
class FaultMode(Enum):
    """FaultMode Enum"""
    COMPARATOR = tc_pb.FaultMode.COMPARATOR
    INTERRUPT = tc_pb.FaultMode.INTERRUPT

@unique
class NoiseFilterMode(Enum):
    """NoiseFilterMode Enum"""
    HZ_60 = tc_pb.NoiseFilterMode.HZ_60
    HZ_50 = tc_pb.NoiseFilterMode.HZ_50

@unique
class AvgMode(Enum):
    """AvgMode Enum"""
    AVG_1 = tc_pb.AvgMode.AVG_1
    AVG_2 = tc_pb.AvgMode.AVG_2
    AVG_4 = tc_pb.AvgMode.AVG_4
    AVG_8 = tc_pb.AvgMode.AVG_8
    AVG_16 = tc_pb.AvgMode.AVG_16

@unique
class TCType(Enum):
    """TCType Enum"""
    TYPE_B = tc_pb.TCType.TYPE_B
    TYPE_E = tc_pb.TCType.TYPE_E
    TYPE_J = tc_pb.TCType.TYPE_J
    TYPE_K = tc_pb.TCType.TYPE_K
    TYPE_N = tc_pb.TCType.TYPE_N
    TYPE_R = tc_pb.TCType.TYPE_R
    TYPE_S = tc_pb.TCType.TYPE_S
    TYPE_T = tc_pb.TCType.TYPE_T

@unique
class VoltageMode(Enum):
    """VoltageMode Enum"""
    GAIN_8 = tc_pb.VoltageMode.GAIN_8
    GAIN_32 = tc_pb.VoltageMode.GAIN_32

@unique
class CJHighMask(Enum):
    """CJHighMask Enum"""
    CJHIGH_MASK_ON = tc_pb.CJHighMask.CJHIGH_MASK_ON
    CJHIGH_MASK_OFF = tc_pb.CJHighMask.CJHIGH_MASK_OFF

@unique
class CJLowMask(Enum):
    """CJLowMask Enum"""
    CJLOW_MASK_ON = tc_pb.CJLowMask.CJLOW_MASK_ON
    CJLOW_MASK_OFF = tc_pb.CJLowMask.CJLOW_MASK_OFF

@unique
class TCHighMask(Enum):
    """TCHighMask Enum"""
    TCHIGH_MASK_ON = tc_pb.TCHighMask.TCHIGH_MASK_ON
    TCHIGH_MASK_OFF = tc_pb.TCHighMask.TCHIGH_MASK_OFF

@unique
class TCLowMask(Enum):
    """TcLowMask Enum"""
    TCLOW_MASK_ON = tc_pb.TCLowMask.TCLOW_MASK_ON
    TCLOW_MASK_OFF = tc_pb.TCLowMask.TCLOW_MASK_OFF

@unique
class OvuvMask(Enum):
    """OvuvMask Enum"""
    OVUV_MASK_ON = tc_pb.OvuvMask.OVUV_MASK_ON
    OVUV_MASK_OFF = tc_pb.OvuvMask.OVUV_MASK_OFF

@unique
class OpenMask(Enum):
    """OpenMask Enum"""
    OPEN_MASK_ON = tc_pb.OpenMask.OPEN_MASK_ON
    OPEN_MASK_OFF = tc_pb.OpenMask.OPEN_MASK_OFF

@unique
class DecBits4(Enum):
    """OpenMask Enum"""
    P0 = tc_pb.DecBits4.P0
    P0_5 = tc_pb.DecBits4.P0_5
    P0_75 = tc_pb.DecBits4.P0_75
    P0_875 = tc_pb.DecBits4.P0_875
    P0_9375 = tc_pb.DecBits4.P0_9375
    P0_4375 = tc_pb.DecBits4.P0_4375
    P0_1875 = tc_pb.DecBits4.P0_1875
    P0_0625 = tc_pb.DecBits4.P0_0625
    P0_5625 = tc_pb.DecBits4.P0_5625
    P0_8125 = tc_pb.DecBits4.P0_8125
    P0_6875 = tc_pb.DecBits4.P0_6875
    P0_25 = tc_pb.DecBits4.P0_25
    P0_125 = tc_pb.DecBits4.P0_125
    P0_625 = tc_pb.DecBits4.P0_625
    P0_3125 = tc_pb.DecBits4.P0_3125
    P0_375 = tc_pb.DecBits4.P0_375

@unique
class DecBits6(Enum):
    "DecBits6 Enum"
    P0 = tc_pb.DecBits6.P0_
    P0_015625 = tc_pb.DecBits6.P0_015625
    P0_03125 = tc_pb.DecBits6.P0_03125
    P0_046875 = tc_pb.DecBits6.P0_046875
    P0_0625 = tc_pb.DecBits6.P0_0625_
    P0_078125 = tc_pb.DecBits6.P0_078125
    P0_09375 = tc_pb.DecBits6.P0_09375
    P0_109375 = tc_pb.DecBits6.P0_109375
    P0_125 = tc_pb.DecBits6.P0_125_
    P0_140625 = tc_pb.DecBits6.P0_140625
    P0_15625 = tc_pb.DecBits6.P0_15625
    P0_171875 = tc_pb.DecBits6.P0_171875
    P0_1875 = tc_pb.DecBits6.P0_1875_
    P0_203125 = tc_pb.DecBits6.P0_203125
    P0_21875 = tc_pb.DecBits6.P0_21875
    P0_234375 = tc_pb.DecBits6.P0_234375
    P0_25 = tc_pb.DecBits6.P0_25_
    P0_265625 = tc_pb.DecBits6.P0_265625
    P0_28125 = tc_pb.DecBits6.P0_28125
    P0_296875 = tc_pb.DecBits6.P0_296875
    P0_3125 = tc_pb.DecBits6.P0_3125_
    P0_328125 = tc_pb.DecBits6.P0_328125
    P0_34375 = tc_pb.DecBits6.P0_34375
    P0_359375 = tc_pb.DecBits6.P0_359375
    P0_375 = tc_pb.DecBits6.P0_375_
    P0_390625 = tc_pb.DecBits6.P0_390625
    P0_40625 = tc_pb.DecBits6.P0_40625
    P0_421875 = tc_pb.DecBits6.P0_421875
    P0_4375 = tc_pb.DecBits6.P0_4375_
    P0_453125 = tc_pb.DecBits6.P0_453125
    P0_46875 = tc_pb.DecBits6.P0_46875
    P0_484375 = tc_pb.DecBits6.P0_484375
    P0_5 = tc_pb.DecBits6.P0_5_
    P0_515625 = tc_pb.DecBits6.P0_515625
    P0_53125 = tc_pb.DecBits6.P0_53125
    P0_546875 = tc_pb.DecBits6.P0_546875
    P0_5625 = tc_pb.DecBits6.P0_5625_
    P0_578125 = tc_pb.DecBits6.P0_578125
    P0_59375 = tc_pb.DecBits6.P0_59375
    P0_609375 = tc_pb.DecBits6.P0_609375
    P0_625 = tc_pb.DecBits6.P0_625_
    P0_640625 = tc_pb.DecBits6.P0_640625
    P0_65625 = tc_pb.DecBits6.P0_65625
    P0_671875 = tc_pb.DecBits6.P0_671875
    P0_6875 = tc_pb.DecBits6.P0_6875_
    P0_703125 = tc_pb.DecBits6.P0_703125
    P0_71875 = tc_pb.DecBits6.P0_71875
    P0_734375 = tc_pb.DecBits6.P0_734375
    P0_75 = tc_pb.DecBits6.P0_75_
    P0_765625 = tc_pb.DecBits6.P0_765625
    P0_78125 = tc_pb.DecBits6.P0_78125
    P0_796875 = tc_pb.DecBits6.P0_796875
    P0_8125 = tc_pb.DecBits6.P0_8125_
    P0_828125 = tc_pb.DecBits6.P0_828125
    P0_84375 = tc_pb.DecBits6.P0_84375
    P0_859375 = tc_pb.DecBits6.P0_859375
    P0_875 = tc_pb.DecBits6.P0_875_
    P0_890625 = tc_pb.DecBits6.P0_890625
    P0_90625 = tc_pb.DecBits6.P0_90625
    P0_921875 = tc_pb.DecBits6.P0_921875
    P0_9375 = tc_pb.DecBits6.P0_9375_
    P0_953125 = tc_pb.DecBits6.P0_953125
    P0_96875 = tc_pb.DecBits6.P0_96875
    P0_984375 = tc_pb.DecBits6.P0_984375



"""Client Enum to Protobuf Enum mapping"""
from enum import Enum, unique
from rpc_generated_protobufs import dac_pb2 as dac_pb

# pylint: disable=no-member
@unique
class DACChannel(Enum):
    """DACChannels Enum"""
    AOUT1 = dac_pb.DACChannels.AOUT1
    AOUT2 = dac_pb.DACChannels.AOUT2
    AOUT3 = dac_pb.DACChannels.AOUT3
    AOUT4 = dac_pb.DACChannels.AOUT4
    AOUT5 = dac_pb.DACChannels.AOUT5
    AOUT6 = dac_pb.DACChannels.AOUT6
    AOUT7 = dac_pb.DACChannels.AOUT7
    AOUT8 = dac_pb.DACChannels.AOUT8



"""Integration test for RPC Client module installation"""
import subprocess
import os
from tempfile import TemporaryDirectory

def test_rpc_client_install():
    """Test client import"""

    # Build the wheel
    subprocess.check_call(["pip3", "install", "build"])
    subprocess.check_call(["python", "-m", "build"])

    # Locate the wheel in the dist directory
    wheel_file = next((file for file in os.listdir("dist") if file.endswith(".whl")), None)
    assert wheel_file, "Wheel file not found"

    with TemporaryDirectory() as tempdir:
        # Install the wheel inside temp directory
        subprocess.check_call(
            ["pip3", "install", os.path.join("dist", wheel_file), "--target", tempdir]
        )

        # Add the temporary directory to sys.path so that we can import from there
        # pylint: disable=import-outside-toplevel
        import sys
        sys.path.insert(0, tempdir)

        # Check imports
        # pylint: disable=import-outside-toplevel, unused-import
        from edgepi_rpc_client.services.led.client_led_service import ClientLEDService
        from edgepi_rpc_client.services.led.led_pb_enums import LEDPins

        # Clean up sys.path
        sys.path.remove(tempdir)



"""Tests server error handling and integration of Rpc Controller with Rpc Handler"""
import logging
import zmq
from pytest import fixture
from rpc_generated_protobufs import rpc_pb2 as rpc_pb, tc_pb2 as tc_pb

_logger = logging.getLogger(__name__)

# pylint: disable= no-member

@fixture(name="socket")
def fixture_test_socket():
    """Returns socket for each test"""
    context = zmq.Context()
    sock = context.socket(zmq.REQ)
    sock.connect('tcp://localhost:5555')
    return sock

def test_bad_rpc_request_data(socket):
    """Tests error handling for data deserialization"""
    # send a bad message
    socket.send(b'BAD PROTO DATA!')
    response = socket.recv()
    # Deserialize rpc response
    response = rpc_pb.RpcResponse.FromString(response)

    assert response.error_code == rpc_pb.ErrorCode.BAD_MESSAGE_DATA

def test_incomplete_rpc_request(socket):
    """Tests error handling for a wrongful rpc request type"""
    # Set up wrongful request
    request = tc_pb.EmptyMsg()
    request = request.SerializeToString()
    # send request
    socket.send(request)
    # Get response
    response = socket.recv()
    response = rpc_pb.RpcResponse.FromString(response)

    assert response.error_code == rpc_pb.ErrorCode.INCOMPLETE_MESSAGE


def test_service_not_found(socket):
    """Test error handling for finding an rpc service"""
    # Create rpc call with bad service name
    request = rpc_pb.RpcRequest(
        service_name="bad service name",
        method_name="_",
        request_proto=b"_"
    )
    # Send
    request = request.SerializeToString()
    socket.send(request)
    # Get response
    response = socket.recv()
    response = rpc_pb.RpcResponse.FromString(response)

    assert response.error_code == rpc_pb.ErrorCode.SERVICE_NOT_FOUND

def test_method_not_found(socket):
    """Test error handling for finding an rpc service method"""
    # Create rpc call with bad method name
    request = rpc_pb.RpcRequest(
        service_name="TcService",
        method_name="bad method name",
        request_proto=b"_"
    )
    # Send
    request = request.SerializeToString()
    socket.send(request)
    # Get response
    response = socket.recv()
    response = rpc_pb.RpcResponse.FromString(response)

    assert response.error_code == rpc_pb.ErrorCode.METHOD_NOT_FOUND

def test_bad_request_proto_data(socket):
    """Test to handle the method input deserialization (request proto within rpc request)"""
    request = rpc_pb.RpcRequest(
        service_name = 'TcService',
        method_name = 'single_sample',
        request_proto = b'bad request proto data'
    )
    # Send
    request = request.SerializeToString()
    socket.send(request)
    # Get response
    response = socket.recv()
    response = rpc_pb.RpcResponse.FromString(response)

    assert response.error_code == rpc_pb.ErrorCode.BAD_MESSAGE_DATA

def test_incomplete_request_proto(socket):
    """Tests error handling for a wrongful method input (request proto within rpc request)"""
    wrong_request_proto = tc_pb.TempReading().SerializeToString()
    request = rpc_pb.RpcRequest(
        service_name = 'TcService',
        method_name = 'set_config',
        request_proto = wrong_request_proto
    )
    # Send
    request = request.SerializeToString()
    socket.send(request)
    # Get response
    response = socket.recv()
    response = rpc_pb.RpcResponse.FromString(response)

    assert response.error_code == rpc_pb.ErrorCode.INCOMPLETE_MESSAGE

def test_reset_error_state(socket):
    """Test that the state of a worker thread's error state is being reset"""

    # Send a bad message

    socket.send(b'BAD PROTO DATA!')
    response = socket.recv()
    # Deserialize rpc response
    response = rpc_pb.RpcResponse.FromString(response)
    _logger.debug("Bad request response: %s", response)
    assert response.error_code == rpc_pb.ErrorCode.BAD_MESSAGE_DATA

    # Send good messages and ensure that no error messages are being returned
    request_proto = tc_pb.EmptyMsg().SerializeToString()
    rpc_request = rpc_pb.RpcRequest(
        service_name = 'TcService',
        method_name = 'single_sample',
        request_proto = request_proto
    )
    request = rpc_request.SerializeToString()\
    # Hammer to hopefully make requests to all threads
    for i in range(10):
        socket.send(request)

        response = socket.recv()
        rpc_response = rpc_pb.RpcResponse.FromString(response)
        _logger.debug("Good request #%s response: %s", i, rpc_response)
        assert not rpc_response.HasField('error_code') and not rpc_response.HasField('error_msg')



"""PWMService integration test"""
import time
import pytest
from edgepi_rpc_client.services.pwm.client_pwm_service import ClientPWMService
from edgepi_rpc_client.services.pwm.pwm_pb_enums import PWMPins, Polarity


@pytest.fixture(name="pwm_service")
def fixture_test_pwm_service():
    """Inits new PWM service client for testing"""
    return ClientPWMService('tcp://localhost:5555')

@pytest.mark.parametrize(
    "args",
    [
        {"pwm_num": PWMPins.PWM1, "polarity": Polarity.NORMAL},
        {"pwm_num": PWMPins.PWM2, "polarity": Polarity.INVERSED},
        {"pwm_num": PWMPins.PWM1, "frequency": 1000},
        {"pwm_num": PWMPins.PWM1, "duty_cycle": 0},
    ]
)
def test_set_config(pwm_service, args):
    """Test for set_config"""
    response_init = pwm_service.init_pwm(pwm_num=args['pwm_num'])
    assert response_init == f"Successfully initialized {args['pwm_num']}."
    response_config = pwm_service.set_config(**args)
    assert response_config == "Successfully applied pwm configurations."

@pytest.mark.parametrize(
    'pwm_num',
    [
        (PWMPins.PWM1),
        (PWMPins.PWM2),
    ]
)
def test_enable(pwm_service, pwm_num):
    """Test for set_enable"""
    pwm_service.init_pwm(pwm_num)
    response = pwm_service.enable(pwm_num)
    assert response == f"Successfully enabled {pwm_num}."

@pytest.mark.parametrize(
    'pwm_num',
    [
        (PWMPins.PWM1),
        (PWMPins.PWM2),
    ]
)
def test_disable(pwm_service, pwm_num):
    """Test for set_disable"""
    pwm_service.init_pwm(pwm_num)
    response = pwm_service.disable(pwm_num)
    assert response == f"Successfully disabled {pwm_num}."

@pytest.mark.parametrize(
    'pwm_num',
    [
        (PWMPins.PWM1),
        (PWMPins.PWM2),
    ]
)
def test_close(pwm_service, pwm_num):
    """Test for close"""
    pwm_service.init_pwm(pwm_num)
    response = pwm_service.close(pwm_num)
    assert response == f"Successfully closed {pwm_num}."


@pytest.mark.parametrize(
    'pwm_num',
    [
        (PWMPins.PWM1),
        (PWMPins.PWM2),
    ]
)
def test_get_frequency(pwm_service, pwm_num):
    """Test for get_frequency"""
    pwm_service.init_pwm(pwm_num)
    frequency = pwm_service.get_frequency(pwm_num)
    assert isinstance(frequency, float)

@pytest.mark.parametrize(
    'pwm_num',
    [
        (PWMPins.PWM1),
        (PWMPins.PWM2),
    ]
)
def test_get_duty_cycle(pwm_service, pwm_num):
    """Test for get_duty_cycle"""
    pwm_service.init_pwm(pwm_num)
    duty_cycle = pwm_service.get_duty_cycle(pwm_num)
    assert isinstance(duty_cycle, float)

@pytest.mark.parametrize(
    'pwm_num',
    [
        (PWMPins.PWM1),
        (PWMPins.PWM2),
    ]
)
def test_get_polarity(pwm_service, pwm_num):
    """Test for get_polarity"""
    pwm_service.init_pwm(pwm_num)
    polarity = pwm_service.get_polarity(pwm_num)
    assert polarity in Polarity

@pytest.mark.parametrize(
    'pwm_num',
    [
        (PWMPins.PWM1),
        (PWMPins.PWM2),
    ]
)
def test_get_enabled(pwm_service, pwm_num):
    """Test for get_enabled"""
    pwm_service.init_pwm(pwm_num)
    enabled = pwm_service.get_enabled(pwm_num)
    assert isinstance(enabled, bool)

def test_with_edgepi(pwm_service):
    """Used for manual testing of PWM functionality."""
    pwm_service.init_pwm(pwm_num=PWMPins.PWM2)
    pwm_service.set_config(pwm_num=PWMPins.PWM2, polarity=Polarity.NORMAL)
    pwm_service.enable(PWMPins.PWM2)
    time.sleep(3)
    pwm_service.set_config(pwm_num=PWMPins.PWM2, polarity=Polarity.INVERSED)
    time.sleep(3)
    pwm_service.set_config(pwm_num=PWMPins.PWM2, polarity=Polarity.NORMAL)
    


"""TcService integration test"""
import logging
import pytest
from edgepi.tc.edgepi_tc import EdgePiTC
from edgepi.tc.tc_constants import TCAddresses, CJMode as sdkCJMode, ConvMode as sdkConvMode
from edgepi_rpc_client.services.tc.client_tc_service import ClientTcService
from edgepi_rpc_client.services.tc.tc_pb_enums import (
    AvgMode,
    CJHighMask,
    CJLowMask,
    CJMode,
    ConvMode,
    DecBits4,
    DecBits6,
    FaultMode,
    NoiseFilterMode,
    OpenCircuitMode,
    OpenMask,
    OvuvMask,
    TCHighMask,
    TCLowMask,
    TCType,
    VoltageMode,
)

_logger = logging.getLogger(__name__)

@pytest.fixture(name="tc_service")
def fixture_test_tc_service():
    """Inits new tc service client for testing"""
    return ClientTcService('tcp://localhost:5555')

@pytest.fixture(name="edgepi_tc")
def fixture_test_edgepi_tc():
    """Inits new EdgePiTc for testing"""
    return EdgePiTC()


@pytest.mark.parametrize(
    "args, set_reg_value, updated_regs",
    [
        (
            {"conversion_mode": ConvMode.SINGLE},
            {TCAddresses.CR0_W.value: 0x80},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"conversion_mode": ConvMode.AUTO},
            None,
            {TCAddresses.CR0_W.value: 0x80},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.HIGH_INPUT_IMPEDANCE},
            None,
            {TCAddresses.CR0_W.value: 0x30},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.MED_INPUT_IMPEDANCE},
            None,
            {TCAddresses.CR0_W.value: 0x20},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.LOW_INPUT_IMPEDANCE},
            None,
            {TCAddresses.CR0_W.value: 0x10},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.DISABLED},
            {TCAddresses.CR0_W.value: 0x10},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"cold_junction_mode": CJMode.DISABLE},
            None,
            {TCAddresses.CR0_W.value: 0x08},
        ),
        (
            {"cold_junction_mode": CJMode.ENABLE},
            {TCAddresses.CR0_W.value: 0x08},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"fault_mode": FaultMode.INTERRUPT},
            None,
            {TCAddresses.CR0_W.value: 0x04},
        ),
        (
            {"fault_mode": FaultMode.COMPARATOR},
            {TCAddresses.CR0_W.value: 0x04},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"noise_filter_mode": NoiseFilterMode.HZ_60},
            {TCAddresses.CR0_W.value: 0x01},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"noise_filter_mode": NoiseFilterMode.HZ_50},
            None,
            {TCAddresses.CR0_W.value: 0x01},
        ),
        (
            {"average_mode": AvgMode.AVG_16},
            None,
            {TCAddresses.CR1_W.value: 0x43},
        ),
        (
            {"average_mode": AvgMode.AVG_8},
            None,
            {TCAddresses.CR1_W.value: 0x33},
        ),
        (
            {"average_mode": AvgMode.AVG_4},
            None,
            {TCAddresses.CR1_W.value: 0x23},
        ),
        (
            {"average_mode": AvgMode.AVG_2},
            None,
            {TCAddresses.CR1_W.value: 0x13},
        ),
        (
            {"average_mode": AvgMode.AVG_1},
            {TCAddresses.CR1_W.value: 0x13},
            {TCAddresses.CR1_W.value: 0x03},
        ),
        (
            {"tc_type": TCType.TYPE_B},
            None,
            {TCAddresses.CR1_W.value: 0x00},
        ),
        (
            {"tc_type": TCType.TYPE_E},
            None,
            {TCAddresses.CR1_W.value: 0x01},
        ),
        (
            {"tc_type": TCType.TYPE_J},
            None,
            {TCAddresses.CR1_W.value: 0x02},
        ),
        (
            {"tc_type": TCType.TYPE_K},
            None,
            {TCAddresses.CR1_W.value: 0x03},
        ),
        (
            {"tc_type": TCType.TYPE_N},
            None,
            {TCAddresses.CR1_W.value: 0x04},
        ),
        (
            {"tc_type": TCType.TYPE_R},
            None,
            {TCAddresses.CR1_W.value: 0x05},
        ),
        (
            {"tc_type": TCType.TYPE_S},
            None,
            {TCAddresses.CR1_W.value: 0x06},
        ),
        (
            {"tc_type": TCType.TYPE_T},
            None,
            {TCAddresses.CR1_W.value: 0x07},
        ),
        (
            {"voltage_mode": VoltageMode.GAIN_8},
            None,
            {TCAddresses.CR1_W.value: 0x08},
        ),
        (
            {"voltage_mode": VoltageMode.GAIN_32},
            None,
            {TCAddresses.CR1_W.value: 0x0C},
        ),
        (
            {"cj_high_mask": CJHighMask.CJHIGH_MASK_OFF},
            None,
            {TCAddresses.MASK_W.value: 0xDF},
        ),
        (
            {"cj_high_mask": CJHighMask.CJHIGH_MASK_ON},
            {TCAddresses.MASK_W.value: 0xDF},
            {TCAddresses.MASK_W.value: 0xFF},
        ),
        (
            {"cj_low_mask": CJLowMask.CJLOW_MASK_OFF},
            None,
            {TCAddresses.MASK_W.value: 0xEF},
        ),
        (
            {"cj_low_mask": CJLowMask.CJLOW_MASK_ON},
            {TCAddresses.MASK_W.value: 0xEF},
            {TCAddresses.MASK_W.value: 0xFF},
        ),
        (
            {"tc_high_mask": TCHighMask.TCHIGH_MASK_OFF},
            None,
            {TCAddresses.MASK_W.value: 0xF7},
        ),
        (
            {"tc_high_mask": TCHighMask.TCHIGH_MASK_ON},
            {TCAddresses.MASK_W.value: 0xF7},
            {TCAddresses.MASK_W.value: 0xFF},
        ),
        (
            {"tc_low_mask": TCLowMask.TCLOW_MASK_OFF},
            None,
            {TCAddresses.MASK_W.value: 0xFB},
        ),
        (
            {"tc_low_mask": TCLowMask.TCLOW_MASK_ON},
            {TCAddresses.MASK_W.value: 0xFB},
            {TCAddresses.MASK_W.value: 0xFF},
        ),
        (
            {"ovuv_mask": OvuvMask.OVUV_MASK_OFF},
            None,
            {TCAddresses.MASK_W.value: 0xFD},
        ),
        (
            {"ovuv_mask": OvuvMask.OVUV_MASK_ON},
            {TCAddresses.MASK_W.value: 0xFD},
            {TCAddresses.MASK_W.value: 0xFF},
        ),
        (
            {"open_mask": OpenMask.OPEN_MASK_OFF},
            None,
            {TCAddresses.MASK_W.value: 0xFE},
        ),
        (
            {"open_mask": OpenMask.OPEN_MASK_ON},
            {TCAddresses.MASK_W.value: 0xFE},
            {TCAddresses.MASK_W.value: 0xFF},
        ),
        (
            {"cj_high_threshold": 100},
            None,
            {TCAddresses.CJHF_W.value: 0x64},
        ),
        ({"cj_low_threshold": -16}, None, {TCAddresses.CJLF_W.value: 0x90}),
        (
            {"lt_high_threshold": 1000, "lt_high_threshold_decimals": DecBits4.P0_9375},
            None,
            {TCAddresses.LTHFTH_W.value: 0x3E, TCAddresses.LTHFTL_W.value: 0x8F},
        ),
        (
            {"lt_low_threshold": -55, "lt_low_threshold_decimals": DecBits4.P0_9375},
            None,
            {TCAddresses.LTLFTH_W.value: 0x83, TCAddresses.LTLFTL_W.value: 0x7F},
        ),
        (
            {"cj_offset": 4, "cj_offset_decimals": DecBits4.P0_9375},
            None,
            {TCAddresses.CJTO_W.value: 0x4F},
        ),
        (
            {"cj_temp": 100, "cj_temp_decimals": DecBits6.P0_984375},
            {TCAddresses.CR0_W.value: 0x08},  # disable cold junction sensor
            {TCAddresses.CJTH_W.value: 0x64, TCAddresses.CJTL_W.value: 0xFC},
        ),
    ],
)

def test_set_config(tc_service,edgepi_tc, args, set_reg_value, updated_regs):
    """Tests set_config tc service method"""

    # reset registers to default values
    edgepi_tc.reset_registers()

    # modify default values for this test only if needed
    if set_reg_value is not None:
        for addx, value in set_reg_value.items():
            # pylint: disable=protected-access
            edgepi_tc._EdgePiTC__write_to_register(addx, value)

    # update registers with user args
    response  = tc_service.set_config(**args)
    assert response == 'Successfully applied thermocouple configurations using set_config'

    # if overwriting cold junction temp, re-enable sensing to return
    # CR0 to default value for value comparison below
    if "cj_temp" in args.keys() or "cj_temp_decimals" in args.keys():
        tc_service.set_config(cold_junction_mode=CJMode.ENABLE)

    # read updated register values. pylint: disable=protected-access
    reg_values = edgepi_tc._EdgePiTC__read_registers_to_map()

    # compare to expected register values
    for addx, value in reg_values.items():
        # these require cold-junction sensor to be disabled, otherwise
        # they are continously updated and cannot be tested against default values
        if addx in (TCAddresses.CJTH_W.value, TCAddresses.CJTL_W.value):
            continue

        # check registers not updated have not been changed
        if addx not in updated_regs:
            assert value == edgepi_tc.default_reg_values[addx]
        # check updates were applied
        else:
            assert value == updated_regs[addx]


def test_single_sample(tc_service):
    """Test for single_sample method"""
    # test passes if both cold junction and linearized temp readings are float
    temps = tc_service.single_sample()
    assert len(temps) == 2
    for temp in temps:
        assert isinstance(temp,float)

def test_read_temperatures(tc_service):
    """Test for read_temperatures method"""
    # Call read_temperatures
    temps = tc_service.read_temperatures()

    _logger.info("%s", temps)
    assert len(temps) == 2
    for temp in temps:
        assert isinstance(temp,float)

@pytest.mark.parametrize(
        'filter_at_fault',
        [
            (True), (False)
        ]
)

def test_read_faults(tc_service, edgepi_tc, filter_at_fault):
    """Test for read_faults method"""

    # Call method
    response =  tc_service.read_faults(filter_at_fault)

    _logger.debug("RESPONSE FAULT DICT: %s", response)

    # Do actual SDK fault reading method
    edgepi_faults = edgepi_tc.read_faults()

    # Compare
    for _,fault in edgepi_faults.items():
        fault_type = str(fault.fault_type)
        assert fault.fault_type in response
        assert response[fault_type]['Fault Type'] == fault_type
        assert response[fault_type]['At Fault'] == fault.at_fault
        assert response[fault_type]['Fault Message'] == str(fault.err_msg.value)
        assert response[fault_type]['Fault Masked'] == fault.is_masked

def test_clear_faults(tc_service):
    """Test for clear_faults method"""
    response = tc_service.clear_faults()

    assert response == 'Successfully called clear_faults'


def test_reset_registers(tc_service):
    """Test for reset_registers method"""
    response = tc_service.reset_registers()

    assert response == 'Successfully called reset_registers'

def test_overwrite_cold_junction_temp(tc_service,edgepi_tc):
    """Test for overwrite cold_junction_temp"""
    # Disable cold junction sensor and set to single conversion mode
    edgepi_tc.set_config(conversion_mode = sdkConvMode.SINGLE, cold_junction_mode=sdkCJMode.DISABLE)
    assert edgepi_tc.tc_state.cold_junction

    # Call method
    response = tc_service.overwrite_cold_junction_temp(
        cj_temp=20, cj_temp_decimals=DecBits6.P0_765625)
    assert response == 'Successfully called overwrite_cold_junction_temp'
    # Verify the write
    cj_temp,_ = edgepi_tc.single_sample()
    assert cj_temp == 20.765625

    # Re-enable cold junction sensor
    edgepi_tc.set_config(cold_junction_mode=sdkCJMode.ENABLE)
    assert not edgepi_tc.tc_state.cold_junction
    # Verify the reset
    cj_temp,_ = edgepi_tc.single_sample()
    assert cj_temp != 20.765625



"""
Client for tc service. Utilizes ClientRpcChannel to send/recieve
and serialize/deserialize messages.
"""
import logging
from rpc_generated_protobufs import tc_pb2 as tc_pb
from edgepi_rpc_client.util.helpers import (
    create_config_request_from_args, filter_arg_values, get_server_response
)
from edgepi_rpc_client.client_rpc_channel import ClientRpcChannel
from edgepi_rpc_client.services.tc.tc_pb_enums import (
    AvgMode,
    CJHighMask,
    CJLowMask,
    CJMode,
    ConvMode,
    DecBits4,
    DecBits6,
    FaultMode,
    NoiseFilterMode,
    OpenCircuitMode,
    OpenMask,
    OvuvMask,
    TCHighMask,
    TCLowMask,
    TCType,
    VoltageMode,
)

_logger = logging.getLogger(__name__)

# pylint: disable=no-member
class ClientTcService():
    """Client Methods for Tc Service"""
    def __init__(self, transport):
        self.client_rpc_channel = ClientRpcChannel(transport)
        self.service_stub = tc_pb.TcService_Stub(self.client_rpc_channel)
        self.rpc_controller = None

    def _generate_faults_dict(self, faults_msg):
        faults_dict = {}

        for fault_msg in faults_msg.fault:
            faults_dict[fault_msg.fault_type] = {
                'Fault Type': fault_msg.fault_type,
                'At Fault': fault_msg.at_fault,
                'Fault Message': fault_msg.err_msg,
                'Fault Masked': fault_msg.is_masked
            }

        return faults_dict

    # pylint: disable=unused-argument, too-many-arguments, too-many-locals
    def set_config(self, conversion_mode: ConvMode = None,
        open_circuit_mode: OpenCircuitMode = None,
        cold_junction_mode: CJMode = None,
        fault_mode: FaultMode = None,
        noise_filter_mode: NoiseFilterMode = None,
        average_mode: AvgMode = None,
        tc_type: TCType = None,
        voltage_mode: VoltageMode = None,
        cj_high_mask: CJHighMask = None,
        cj_low_mask: CJLowMask = None,
        tc_high_mask: TCHighMask = None,
        tc_low_mask: TCLowMask = None,
        ovuv_mask: OvuvMask = None,
        open_mask: OpenMask = None,
        cj_high_threshold: int = None,
        cj_low_threshold: int = None,
        lt_high_threshold: int = None,
        lt_high_threshold_decimals: DecBits4 = None,
        lt_low_threshold: int = None,
        lt_low_threshold_decimals: DecBits4 = None,
        cj_offset: int = None,
        cj_offset_decimals: DecBits4 = None,
        cj_temp: int = None,
        cj_temp_decimals: DecBits6 = None,
    ):
        """set_config method for sdk tc module."""
        # Get a dictionary of arguments that are not None.
        config_args_dict= filter_arg_values(locals(), 'self', None)
        _logger.debug("Config argument dictionary: %s", config_args_dict)

        # Create config request
        config_msg = tc_pb.Config()
        arg_msg = tc_pb.Config().ConfArg()
        request = create_config_request_from_args(config_msg, arg_msg, config_args_dict)

        # Call the SDK method through the rpc channel client
        rpc_response = self.service_stub.set_config(self.rpc_controller,request)

        response = get_server_response(rpc_response, tc_pb.SuccessMsg)

        return response.content

    def single_sample(self):
        """single_sample method for sdk tc module"""
        request = tc_pb.EmptyMsg()
        # call the SDK method through rpc channel client
        rpc_response = self.service_stub.single_sample(self.rpc_controller,request)

        response = get_server_response(rpc_response, tc_pb.TempReading)

        temps = (response.cj_temp, response.lin_temp)

        return temps

    def read_temperatures(self):
        """read_temperatures method for sdk tc module"""
        request = tc_pb.EmptyMsg()
        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.read_temperatures(self.rpc_controller,request)

        response = get_server_response(rpc_response, tc_pb.TempReading)

        temps = (response.cj_temp, response.lin_temp)

        return temps

    def read_faults(self,filter_at_fault = True):
        """read_faults method for sdk tc module"""
        request = tc_pb.FilterFaults()
        request.filter_at_fault = filter_at_fault

        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.read_faults(self.rpc_controller,request)

        response = get_server_response(rpc_response, tc_pb.Faults)

        result_dict = self._generate_faults_dict(response)

        return result_dict

    def clear_faults(self):
        """clear_faults method for sdk tc module"""
        request = tc_pb.EmptyMsg()

        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.clear_faults(self.rpc_controller,request)

        response = get_server_response(rpc_response, tc_pb.SuccessMsg)

        return response.content

    def reset_registers(self):
        """reset_registers method for sdk tc module"""
        request = tc_pb.EmptyMsg()

        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.reset_registers(self.rpc_controller,request)

        response = get_server_response(rpc_response, tc_pb.SuccessMsg)

        return response.content

    def overwrite_cold_junction_temp(self, cj_temp: int, cj_temp_decimals: DecBits6):
        """overwrite_cold_junction_temp for sdk tc module"""
        request = tc_pb.CJtemp(
            cj_temp = cj_temp,
            cj_temp_decimals = cj_temp_decimals.value
        )

        # Call SDK method through rpc channel client
        rpc_response = self.service_stub.overwrite_cold_junction_temp(self.rpc_controller,request)

        response = get_server_response(rpc_response, tc_pb.SuccessMsg)

        return response.content
